@@ -1,3 +1,4 @@
+#
 # MIT License
 #
 # Copyright (c) 2020 Airbyte
@@ -19,19 +20,23 @@
 # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 # SOFTWARE.
+#
 
 
 from abc import ABC, abstractmethod
-from typing import Any, Iterable, List, Mapping, MutableMapping, Optional
+from typing import Any, Iterable, List, Mapping, MutableMapping, Optional, Union
 
 import requests
 from airbyte_cdk.models import SyncMode
 from airbyte_cdk.sources.streams.core import Stream
 
 from .auth.core import HttpAuthenticator, NoAuth
-from .exceptions import DefaultBackoffException, UserDefinedBackoffException
+from .exceptions import DefaultBackoffException, RequestBodyException, UserDefinedBackoffException
 from .rate_limiting import default_backoff_handler, user_defined_backoff_handler
 
+# list of all possible HTTP methods which can be used for sending of request bodies
+BODY_REQUEST_METHODS = ("POST", "PUT", "PATCH")
+
 
 class HttpStream(Stream, ABC):
     """
@@ -54,7 +59,7 @@ class HttpStream(Stream, ABC):
     @property
     def http_method(self) -> str:
         """
-        Override if needed. See get_request_data if using POST.
+        Override if needed. See get_request_data/get_request_json if using POST/PUT/PATCH.
         """
         return "GET"
 
@@ -104,6 +109,23 @@ class HttpStream(Stream, ABC):
         """
         return {}
 
+    def request_body_data(
+        self,
+        stream_state: Mapping[str, Any],
+        stream_slice: Mapping[str, Any] = None,
+        next_page_token: Mapping[str, Any] = None,
+    ) -> Optional[Union[Mapping, str]]:
+        """
+        Override when creating POST/PUT/PATCH requests to populate the body of the request with a non-JSON payload.
+
+        If returns a ready text that it will be sent as is.
+        If returns a dict that it will be converted to a urlencoded form.
+           E.g. {"key1": "value1", "key2": "value2"} => "key1=value1&key2=value2"
+
+        At the same time only one of the 'request_body_data' and 'request_body_json' functions can be overridden.
+        """
+        return None
+
     def request_body_json(
         self,
         stream_state: Mapping[str, Any],
@@ -111,11 +133,25 @@ class HttpStream(Stream, ABC):
         next_page_token: Mapping[str, Any] = None,
     ) -> Optional[Mapping]:
         """
-        TODO make this possible to do for non-JSON APIs
-        Override when creating POST requests to populate the body of the request with a JSON payload.
+        Override when creating POST/PUT/PATCH requests to populate the body of the request with a JSON payload.
+
+        At the same time only one of the 'request_body_data' and 'request_body_json' functions can be overridden.
         """
         return None
 
+    def request_kwargs(
+        self,
+        stream_state: Mapping[str, Any],
+        stream_slice: Mapping[str, Any] = None,
+        next_page_token: Mapping[str, Any] = None,
+    ) -> Mapping[str, Any]:
+        """
+        Override to return a mapping of keyword arguments to be used when creating the HTTP request.
+        Any option listed in https://docs.python-requests.org/en/latest/api/#requests.adapters.BaseAdapter.send for can be returned from
+        this method. Note that these options do not conflict with request-level options such as headers, request params, etc..
+        """
+        return {}
+
     @abstractmethod
     def parse_response(
         self,
@@ -156,21 +192,26 @@ class HttpStream(Stream, ABC):
         return None
 
     def _create_prepared_request(
-        self, path: str, headers: Mapping = None, params: Mapping = None, json: Any = None
+        self, path: str, headers: Mapping = None, params: Mapping = None, json: Any = None, data: Any = None
     ) -> requests.PreparedRequest:
         args = {"method": self.http_method, "url": self.url_base + path, "headers": headers, "params": params}
+        if self.http_method.upper() in BODY_REQUEST_METHODS:
+            if json and data:
+                raise RequestBodyException(
+                    "At the same time only one of the 'request_body_data' and 'request_body_json' functions can return data"
+                )
+            elif json:
+                args["json"] = json
+            elif data:
+                args["data"] = data
 
-        if self.http_method.upper() == "POST":
-            # TODO support non-json bodies
-            args["json"] = json
-
-        return requests.Request(**args).prepare()
+        return self._session.prepare_request(requests.Request(**args))
 
     # TODO allow configuring these parameters. If we can get this into the requests library, then we can do it without the ugly exception hacks
     #  see https://github.com/litl/backoff/pull/122
     @default_backoff_handler(max_tries=5, factor=5)
     @user_defined_backoff_handler(max_tries=5)
-    def _send_request(self, request: requests.PreparedRequest) -> requests.Response:
+    def _send_request(self, request: requests.PreparedRequest, request_kwargs: Mapping[str, Any]) -> requests.Response:
         """
         Wraps sending the request in rate limit and error handlers.
 
@@ -188,9 +229,8 @@ class HttpStream(Stream, ABC):
         Unexpected transient exceptions use the default backoff parameters.
         Unexpected persistent exceptions are not handled and will cause the sync to fail.
         """
-        response: requests.Response = self._session.send(request)
+        response: requests.Response = self._session.send(request, **request_kwargs)
         if self.should_retry(response):
-
             custom_backoff_time = self.backoff_time(response)
             if custom_backoff_time:
                 raise UserDefinedBackoffException(backoff=custom_backoff_time, request=request, response=response)
@@ -213,17 +253,18 @@ class HttpStream(Stream, ABC):
         stream_state = stream_state or {}
         pagination_complete = False
 
+        next_page_token = None
         while not pagination_complete:
-            next_page_token = None
             request_headers = self.request_headers(stream_state=stream_state, stream_slice=stream_slice, next_page_token=next_page_token)
             request = self._create_prepared_request(
                 path=self.path(stream_state=stream_state, stream_slice=stream_slice, next_page_token=next_page_token),
                 headers=dict(request_headers, **self.authenticator.get_auth_header()),
                 params=self.request_params(stream_state=stream_state, stream_slice=stream_slice, next_page_token=next_page_token),
                 json=self.request_body_json(stream_state=stream_state, stream_slice=stream_slice, next_page_token=next_page_token),
+                data=self.request_body_data(stream_state=stream_state, stream_slice=stream_slice, next_page_token=next_page_token),
             )
-
-            response = self._send_request(request)
+            request_kwargs = self.request_kwargs(stream_state=stream_state, stream_slice=stream_slice, next_page_token=next_page_token)
+            response = self._send_request(request, request_kwargs)
             yield from self.parse_response(response, stream_state=stream_state, stream_slice=stream_slice)
 
             next_page_token = self.next_page_token(response)
